@@ -24,7 +24,7 @@ Each Instruction type acts on a number of :attr:`~Instruction.qubits`, and expec
 Instructions
 ============
 
-We currently support three native instruction types:
+We currently support the following native instruction types:
 
 ================ =========== ====================================== ===========
 name             # of qubits args                                   description
@@ -32,9 +32,15 @@ name             # of qubits args                                   description
 measurement      >= 1        ``key: str``                           Measurement in the Z basis.
 phased_rx        1           ``angle_t: float``, ``phase_t: float`` Phased x-rotation gate.
 cz               2                                                  Controlled-Z gate.
-barrier          >= 2                                               Barrier instruction.
+barrier          >= 2                                               Execution barrier.
 ================ =========== ====================================== ===========
 
+Instructions can be further specified by adding an ``implementation`` field with
+a name for the implementation of the instruction. The default value for this field is
+an empty string which selects the default implementation.
+Support for multiple implementations is currently experimental and in normal use the
+field should be omitted. This selects the default implementation for the instruction.
+
 Measurement
 -----------
 
@@ -44,7 +50,10 @@ All the measurement keys in a circuit must be unique.
 Each qubit may only be measured once.
 The measurement must be the last operation on each qubit, i.e. it cannot be followed by gates.
 
-Example: ``Instruction(name='measurement', qubits=('alice', 'bob', 'charlie'), args={'key': 'm1'})``
+.. code-block:: python
+   :caption: Example
+
+   Instruction(name='measurement', qubits=('alice', 'bob', 'charlie'), args={'key': 'm1'})
 
 
 Phased Rx
@@ -62,7 +71,10 @@ The gate is represented in the standard computational basis by the matrix
 where :math:`\theta` = ``angle_t``, :math:`\phi` = ``phase_t``,
 and :math:`X` and :math:`Y` are Pauli matrices.
 
-Example: ``Instruction(name='phased_rx', qubits=('bob'), args={'angle_t': 0.7, 'phase_t': 0.25})``
+.. code-block:: python
+   :caption: Example
+
+   Instruction(name='phased_rx', qubits=('bob',), args={'angle_t': 0.7, 'phase_t': 0.25})
 
 
 CZ
@@ -72,25 +84,35 @@ Controlled-Z gate. Represented in the standard computational basis by the matrix
 
 .. math:: \text{CZ} = \text{diag}(1, 1, 1, -1).
 
-Symmetric wrt. the qubits it's acting on. Takes no arguments.
+It is symmetric wrt. the qubits it's acting on, and takes no arguments.
 
-Example: ``Instruction(name='cz', qubits=('alice', 'bob'), args={})``
+.. code-block:: python
+   :caption: Example
+
+   Instruction(name='cz', qubits=('alice', 'bob'), args={})
 
 
 Barrier
 -------
 
-Barriers ensure that all operations after the barrier on the qubit subsystems spanned by
-the barrier are only executed when all the operations before the barrier have been completed.
+A barrier instruction affects the physical execution order of the instructions elsewhere in the
+circuit that act on qubits spanned by the barrier.
+It ensures that any such instructions that succeed the barrier are only executed after
+all such instructions that precede the barrier have been completed.
+Hence it can be used to guarantee a specific causal order for the other instructions.
+It takes no arguments, and has no other effect.
 
-Example: ``Instruction(name='barrier', qubits=('alice', 'bob'), args={})``
+.. code-block:: python
+   :caption: Example
+
+   Instruction(name='barrier', qubits=('alice', 'bob'), args={})
 
 
 Circuit output
 ==============
 
-The :class:`RunResult` class represents the results of the quantum circuit execution.
-If the run succeeded, :attr:`RunResult.measurements` contains the output of the batch of circuits,
+The :class:`RunResult` class represents the results of the quantum circuit execution job.
+If the job succeeded, :attr:`RunResult.measurements` contains the output of the batch of circuits,
 consisting of the results of the measurement operations in each circuit.
 It is a list of dictionaries, where each dict maps each measurement key to a 2D array of measurement
 results, represented as a nested list.
@@ -112,7 +134,7 @@ import json
 import os
 from posixpath import join
 import time
-from typing import Any, Optional, Union
+from typing import Any, Callable, Optional, Union
 from uuid import UUID
 import warnings
 
@@ -122,7 +144,7 @@ import requests
 REQUESTS_TIMEOUT = 60
 
 DEFAULT_TIMEOUT_SECONDS = 900
-SECONDS_BETWEEN_CALLS = 3
+SECONDS_BETWEEN_CALLS = float(os.environ.get('IQM_CLIENT_SECONDS_BETWEEN_CALLS', 1.0))
 REFRESH_MARGIN_SECONDS = REQUESTS_TIMEOUT
 
 AUTH_CLIENT_ID = 'iqm_client'
@@ -142,12 +164,12 @@ class CircuitExecutionError(RuntimeError):
 
 
 class APITimeoutError(CircuitExecutionError):
-    """Exception for when executing a task on the server takes too long."""
+    """Exception for when executing a job on the server takes too long."""
 
 
 class Status(str, Enum):
     """
-    Status of a task.
+    Status of a job.
     """
 
     PENDING = 'pending'
@@ -158,29 +180,25 @@ class Status(str, Enum):
 class Instruction(BaseModel):
     """An instruction in a quantum circuit."""
 
-    name: str = Field(..., description='name of the quantum operation', example='measurement')
+    name: str = Field(..., example='measurement')
     """name of the quantum operation"""
-    qubits: tuple[str, ...] = Field(
-        ...,
-        description='names of the logical qubits the operation acts on',
-        example=['alice'],
-    )
+    implementation: Optional[str] = Field(None)
+    """name of the implementation, for experimental use only"""
+    qubits: tuple[str, ...] = Field(..., example=('alice',))
     """names of the logical qubits the operation acts on"""
-    args: dict[str, Any] = Field(
-        ...,
-        description='arguments for the operation',
-        example={'key': 'm'},
-    )
+    args: dict[str, Any] = Field(..., example={'key': 'm'})
     """arguments for the operation"""
 
 
 class Circuit(BaseModel):
     """Quantum circuit to be executed."""
 
-    name: str = Field(..., description='name of the circuit', example='test circuit')
+    name: str = Field(..., example='test circuit')
     """name of the circuit"""
-    instructions: tuple[Instruction, ...] = Field(..., description='instructions comprising the circuit')
+    instructions: tuple[Instruction, ...] = Field(...)
     """instructions comprising the circuit"""
+    metadata: Optional[dict[str, Any]] = Field(None)
+    """arbitrary metadata associated with the circuit"""
 
     def all_qubits(self) -> set[str]:
         """Return the names of all qubits in the circuit."""
@@ -197,9 +215,9 @@ CircuitBatch = list[Circuit]
 class SingleQubitMapping(BaseModel):
     """Mapping of a logical qubit name to a physical qubit name."""
 
-    logical_name: str = Field(..., description='logical qubit name', example='alice')
+    logical_name: str = Field(..., example='alice')
     """logical qubit name"""
-    physical_name: str = Field(..., description='physical qubit name', example='QB1')
+    physical_name: str = Field(..., example='QB1')
     """physical qubit name"""
 
 
@@ -221,30 +239,21 @@ def serialize_qubit_mapping(qubit_mapping: dict[str, str]) -> list[SingleQubitMa
 
 
 class RunRequest(BaseModel):
-    """Request for an IQM quantum computer to execute a batch of quantum circuits.
+    """Request for an IQM quantum computer to run a job that executes a batch of quantum circuits.
 
     Note: all circuits in a batch must measure the same qubits otherwise batch execution fails.
     """
 
-    circuits: CircuitBatch = Field(..., description='batch of quantum circuit(s) to execute')
+    circuits: CircuitBatch = Field(...)
     """batch of quantum circuit(s) to execute"""
-    custom_settings: dict[str, Any] = Field(
-        None,
-        description="""Custom settings to overwrite default IQM hardware settings and calibration data.
-Note: This field should be always None in normal use.""",
-    )
-    """Custom settings to overwrite default IQM hardware settings and calibration data.
+    custom_settings: dict[str, Any] = Field(None)
+    """Custom settings to override default IQM hardware settings and calibration data.
 Note: This field should be always None in normal use."""
-    calibration_set_id: Optional[int] = Field(
-        None, description='ID of the calibration set to use, or None to use the latest calibration set'
-    )
+    calibration_set_id: Optional[int] = Field(None)
     """ID of the calibration set to use, or None to use the latest calibration set"""
-    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
-        None,
-        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names',
-    )
+    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(None)
     """mapping of logical qubit names to physical qubit names, or None if using physical qubit names"""
-    shots: int = Field(..., description='how many times to execute each circuit in the batch')
+    shots: int = Field(...)
     """how many times to execute each circuit in the batch"""
 
 
@@ -259,40 +268,31 @@ CircuitMeasurementResultsBatch = list[CircuitMeasurementResults]
 
 
 class Metadata(BaseModel):
-    """Metadata belonging to a job sumission"""
+    """Metadata describing a circuit execution job."""
 
-    shots: int = Field(..., description='how many times to execute each circuit in the batch')
-    """how many times to execute each circuit in the batch"""
-    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
-        None,
-        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names',
-    )
-    """mapping of logical qubit names to physical qubit names, or None if using physical qubit names"""
-    circuits: CircuitBatch = Field(..., description='batch of quantum circuit(s) to execute')
-    """batch of quantum circuit(s) to execute"""
-    calibration_set_id: Optional[int] = Field(None, description='ID of the calibration set used')
+    calibration_set_id: Optional[int] = Field(None)
     """ID of the calibration set used"""
+    request: RunRequest = Field(...)
+    """copy of the original RunRequest sent to the server"""
 
 
 class RunResult(BaseModel):
-    """Results of executing a batch of circuit(s).
+    """Results of a circuit execution job.
 
     * ``measurements`` is present iff the status is ``'ready'``.
     * ``message`` carries additional information for the ``'failed'`` status.
     * If the status is ``'pending'``, ``measurements`` and ``message`` are ``None``.
     """
 
-    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
-    """current status of the run, in ``{'pending', 'ready', 'failed'}``"""
-    measurements: Optional[CircuitMeasurementResultsBatch] = Field(
-        None, description='if the run has finished successfully, the measurement results for the circuit(s)'
-    )
-    """if the run has finished successfully, the measurement results for the circuit(s)"""
-    message: Optional[str] = Field(None, description='if the run failed, an error message')
-    """if the run failed, an error message"""
-    metadata: Metadata = Field(..., description='metadata about the underlying job request')
-    """metadata about the underlying job request"""
-    warnings: Optional[list[str]] = Field(None, description='list of warning messages')
+    status: Status = Field(...)
+    """current status of the job, in ``{'pending', 'ready', 'failed'}``"""
+    measurements: Optional[CircuitMeasurementResultsBatch] = Field(None)
+    """if the job has finished successfully, the measurement results for the circuit(s)"""
+    message: Optional[str] = Field(None)
+    """if the job failed, an error message"""
+    metadata: Metadata = Field(...)
+    """metadata about the job"""
+    warnings: Optional[list[str]] = Field(None)
     """list of warning messages"""
 
     @staticmethod
@@ -303,7 +303,7 @@ class RunResult(BaseModel):
             inp: value to parse, has to map to RunResult
 
         Returns:
-            parsed run result
+            parsed job result
 
         """
         input_copy = inp.copy()
@@ -311,13 +311,13 @@ class RunResult(BaseModel):
 
 
 class RunStatus(BaseModel):
-    """Status of a circuit execution request."""
+    """Status of a circuit execution job."""
 
-    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
-    """current status of the run, in ``{'pending', 'ready', 'failed'}``"""
-    message: Optional[str] = Field(None, description='if the run failed, an error message')
-    """if the run failed, an error message"""
-    warnings: Optional[list[str]] = Field(None, description='list of warning messages')
+    status: Status = Field(...)
+    """current status of the job, in ``{'pending', 'ready', 'failed'}``"""
+    message: Optional[str] = Field(None)
+    """if the job failed, an error message"""
+    warnings: Optional[list[str]] = Field(None)
     """list of warning messages"""
 
     @staticmethod
@@ -328,7 +328,7 @@ class RunStatus(BaseModel):
             inp: value to parse, has to map to RunResult
 
         Returns:
-            parsed run result
+            parsed job status
 
         """
         input_copy = inp.copy()
@@ -338,22 +338,20 @@ class RunStatus(BaseModel):
 class QuantumArchitectureSpecification(BaseModel):
     """Quantum architecture specification."""
 
-    name: str = Field(..., description='name of the quantum architecture')
+    name: str = Field(...)
     """name of the quantum architecture"""
-    operations: list[str] = Field(..., description='list of operations supported by this quantum architecture')
+    operations: list[str] = Field(...)
     """list of operations supported by this quantum architecture"""
-    qubits: list[str] = Field(..., description='list of qubits of this quantum architecture')
+    qubits: list[str] = Field(...)
     """list of qubits of this quantum architecture"""
-    qubit_connectivity: list[list[str]] = Field(..., description='qubit connectivity of this quantum architecture')
+    qubit_connectivity: list[list[str]] = Field(...)
     """qubit connectivity of this quantum architecture"""
 
 
 class QuantumArchitecture(BaseModel):
     """Quantum architecture as returned by Cortex."""
 
-    quantum_architecture: QuantumArchitectureSpecification = Field(
-        ..., description='details about the quantum architecture'
-    )
+    quantum_architecture: QuantumArchitectureSpecification = Field(...)
     """details about the quantum architecture"""
 
 
@@ -377,19 +375,15 @@ class AuthRequest(BaseModel):
 
     """
 
-    client_id: str = Field(..., description='name of the client for all request types')
+    client_id: str = Field(...)
     """name of the client for all request types"""
-    grant_type: Optional[GrantType] = Field(
-        None, description="type of token request, in ``{'password', 'refresh_token'}``"
-    )
+    grant_type: Optional[GrantType] = Field(None)
     """type of token request, in ``{'password', 'refresh_token'}``"""
-    username: Optional[str] = Field(None, description="username for grant type ``'password'``")
+    username: Optional[str] = Field(None)
     """username for grant type ``'password'``"""
-    password: Optional[str] = Field(None, description="password for grant type ``'password'``")
+    password: Optional[str] = Field(None)
     """password for grant type ``'password'``"""
-    refresh_token: Optional[str] = Field(
-        None, description="refresh token for grant type ``'refresh_token'`` and logout request"
-    )
+    refresh_token: Optional[str] = Field(None)
     """refresh token for grant type ``'refresh_token'`` and logout request"""
 
 
@@ -401,15 +395,15 @@ class Credentials(BaseModel):
       refreshed periodically.
     """
 
-    auth_server_url: str = Field(..., description='Base URL of the authentication server')
+    auth_server_url: str = Field(...)
     """Base URL of the authentication server"""
-    username: str = Field(..., description='username for logging in to the server')
+    username: str = Field(...)
     """username for logging in to the server"""
-    password: str = Field(..., description='password for logging in to the server')
+    password: str = Field(...)
     """password for logging in to the server"""
-    access_token: Optional[str] = Field(None, description='current access token of the session')
+    access_token: Optional[str] = Field(None)
     """current access token of the session"""
-    refresh_token: Optional[str] = Field(None, description='current refresh token of the session')
+    refresh_token: Optional[str] = Field(None)
     """current refresh token of the session"""
 
 
@@ -420,9 +414,9 @@ class ExternalToken(BaseModel):
       external resource, e.g. file generated by Cortex CLI's token manager.
     """
 
-    auth_server_url: str = Field(..., description='Base URL of the authentication server')
+    auth_server_url: str = Field(...)
     """Base URL of the authentication server"""
-    access_token: str = Field(None, description='current access token of the session')
+    access_token: str = Field(None)
     """current access token of the session"""
 
 
@@ -503,13 +497,13 @@ class IQMClient:
             must not be set.
 
     Keyword Args:
-        auth_server_url: Optional base URL of the authentication server.
+        auth_server_url (str): Optional base URL of the authentication server.
             This can also be set in the IQM_AUTH_SERVER environment variable.
             If unset, requests will be sent unauthenticated.
-        username: Optional username to log in to authentication server.
+        username (str): Optional username to log in to authentication server.
             This can also be set in the IQM_AUTH_USERNAME environment variable.
             Username must be set if ``auth_server_url`` is set.
-        password: Optional password to log in to authentication server.
+        password (str): Optional password to log in to authentication server.
             This can also be set in the IQM_AUTH_PASSWORD environment variable.
             Password must be set if ``auth_server_url`` is set.
     """
@@ -522,6 +516,7 @@ class IQMClient:
         if tokens_file and credentials:
             raise ClientConfigurationError('Either external token or credentials must be provided. Both were provided.')
         self._base_url = url
+        self._tokens_file = tokens_file
         self._external_token = _get_external_token(tokens_file)
         if not self._external_token:
             self._credentials = _get_credentials(credentials)
@@ -536,6 +531,20 @@ class IQMClient:
         except Exception:  # pylint: disable=broad-except
             pass
 
+    def _retry_request_on_error(self, request: Callable[[], requests.Response]) -> requests.Response:
+        """This is a temporary workaround for 502 errors.
+        The current implementation of the server side can run out of network connections
+        and silently drop incoming connections making IQM Client to fail with 502 errors."""
+
+        while True:
+            result = request()
+            if result.status_code == 502:
+                time.sleep(SECONDS_BETWEEN_CALLS)
+                continue
+            break
+
+        return result
+
     # pylint: disable=too-many-locals
     def submit_circuits(
         self,
@@ -549,17 +558,17 @@ class IQMClient:
         """Submits a batch of quantum circuits for execution on a quantum computer.
 
         Args:
-            circuits: list of circuit to be executed
-            qubit_mapping: Mapping of human-readable (logical) qubit names in to physical qubit names.
+            circuits: list of circuits to be executed
+            qubit_mapping: Mapping of logical qubit names to physical qubit names.
                 Can be set to ``None`` if all ``circuits`` already use physical qubit names.
                 Note that the ``qubit_mapping`` is used for all ``circuits``.
-            custom_settings: custom settings to overwrite default settings and calibration data.
-                Note: This field should be always None in normal use.
-            calibration_set_id: ID of the calibration set to use instead of ``settings``
-            shots: number of times ``circuit`` is executed
+            custom_settings: Custom settings to override default settings and calibration data.
+                Note: This field should always be ``None`` in normal use.
+            calibration_set_id: ID of the calibration set to use, or ``None`` to use the latest one
+            shots: number of times ``circuits`` are executed
 
         Returns:
-            ID for the created task. This ID is needed to query the status and the execution results.
+            ID for the created job. This ID is needed to query the job status and the execution results.
         """
         serialized_qubit_mapping: Optional[list[SingleQubitMapping]] = None
         if qubit_mapping is not None:
@@ -594,35 +603,55 @@ class IQMClient:
         if bearer_token:
             headers['Authorization'] = bearer_token
 
-        result = requests.post(
-            join(self._base_url, 'jobs'), json=data.dict(exclude_none=True), headers=headers, timeout=REQUESTS_TIMEOUT
+        try:
+            # check if someone is trying to profile us with OpenTelemetry
+            # pylint: disable=import-outside-toplevel
+            # pylint: disable=import-error
+            from opentelemetry import propagate
+
+            propagate.inject(headers)
+        except ImportError as _:
+            # no OpenTelemetry, no problem
+            pass
+
+        result = self._retry_request_on_error(
+            lambda: requests.post(
+                join(self._base_url, 'jobs'),
+                json=data.dict(exclude_none=True),
+                headers=headers,
+                timeout=REQUESTS_TIMEOUT,
+            )
         )
 
         if result.status_code == 401:
-            raise ClientConfigurationError(f'Authentication failed: {result.text}')
+            raise ClientAuthenticationError(f'Authentication failed: {result.text}')
 
         result.raise_for_status()
         return UUID(result.json()['id'])
 
     def get_run(self, job_id: UUID) -> RunResult:
-        """Query the status and results of the running task.
+        """Query the status and results of a submitted job.
 
         Args:
-            job_id: id of the task
+            job_id: id of the job to query
 
         Returns:
-            result of the run (can be pending)
+            result of the job (can be pending)
 
         Raises:
             HTTPException: http exceptions
             CircuitExecutionError: IQM server specific exceptions
         """
         bearer_token = self._get_bearer_token()
-        result = requests.get(
-            join(self._base_url, 'jobs/', str(job_id)),
-            headers=None if not bearer_token else {'Authorization': bearer_token},
-            timeout=REQUESTS_TIMEOUT,
+
+        result = self._retry_request_on_error(
+            lambda: requests.get(
+                join(self._base_url, 'jobs/', str(job_id)),
+                headers=None if not bearer_token else {'Authorization': bearer_token},
+                timeout=REQUESTS_TIMEOUT,
+            )
         )
+
         result.raise_for_status()
         run_result = RunResult.from_dict(result.json())
         if run_result.warnings:
@@ -633,24 +662,28 @@ class IQMClient:
         return run_result
 
     def get_run_status(self, job_id: UUID) -> RunStatus:
-        """Query the status of the running task.
+        """Query the status of a submitted job.
 
         Args:
-            job_id: id of the task
+            job_id: id of the job to query
 
         Returns:
-            status of the run
+            status of the job
 
         Raises:
             HTTPException: http exceptions
             CircuitExecutionError: IQM server specific exceptions
         """
         bearer_token = self._get_bearer_token()
-        result = requests.get(
-            join(self._base_url, 'jobs/', str(job_id), 'status'),
-            headers=None if not bearer_token else {'Authorization': bearer_token},
-            timeout=REQUESTS_TIMEOUT,
+
+        result = self._retry_request_on_error(
+            lambda: requests.get(
+                join(self._base_url, 'jobs/', str(job_id), 'status'),
+                headers=None if not bearer_token else {'Authorization': bearer_token},
+                timeout=REQUESTS_TIMEOUT,
+            )
         )
+
         result.raise_for_status()
         run_result = RunStatus.from_dict(result.json())
         if run_result.warnings:
@@ -659,14 +692,17 @@ class IQMClient:
         return run_result
 
     def wait_for_results(self, job_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
-        """Poll results until run is ready, failed, or timed out.
+        """Poll results until a job is either ready, failed, or timed out.
+           Note, that jobs handling on the server side is async and if we try to request the results
+           right after submitting the job (which is usually the case)
+           we will find the job is still pending at least for the first query.
 
         Args:
-            job_id: id of the task to wait
+            job_id: id of the job to wait for
             timeout_secs: how long to wait for a response before raising an APITimeoutError
 
         Returns:
-            run result
+            job result
 
         Raises:
             APITimeoutError: time exceeded the set timeout
@@ -677,7 +713,7 @@ class IQMClient:
             if results.status != Status.PENDING:
                 return results
             time.sleep(SECONDS_BETWEEN_CALLS)
-        raise APITimeoutError(f"The task didn't finish in {timeout_secs} seconds.")
+        raise APITimeoutError(f"The job didn't finish in {timeout_secs} seconds.")
 
     def get_quantum_architecture(self) -> QuantumArchitectureSpecification:
         """Retrieve quantum architecture from Cortex.
@@ -687,6 +723,7 @@ class IQMClient:
 
         Raises:
             APITimeoutError: time exceeded the set timeout
+            ClientConfigurationError: if no valid authentication is provided
         """
         bearer_token = self._get_bearer_token()
         result = requests.get(
@@ -694,6 +731,16 @@ class IQMClient:
             headers=None if not bearer_token else {'Authorization': bearer_token},
             timeout=REQUESTS_TIMEOUT,
         )
+
+        # /quantum_architecture is not a strictly authenticated endpoint,
+        # so we need to handle 302 redirects to the auth server login page
+        if result.history and any(
+            response.status_code == 302 for response in result.history
+        ):  # pragma: no cover (generators are broken in coverage)
+            raise ClientConfigurationError('Authentication is required.')
+        if result.status_code == 401:
+            raise ClientAuthenticationError(f'Authentication failed: {result.text}')
+
         result.raise_for_status()
         return QuantumArchitecture(**result.json()).quantum_architecture
 
@@ -739,6 +786,11 @@ class IQMClient:
             is not available.
         """
         if self._external_token:
+            # If access token obtained from external tokens file expires soon, get updated token from the tokens file
+            if _time_left_seconds(self._external_token.access_token) < REFRESH_MARGIN_SECONDS:
+                self._external_token = _get_external_token(self._tokens_file)
+                if not self._external_token:
+                    return None
             return f'Bearer {self._external_token.access_token}'
         if self._credentials is None or not self._credentials.access_token:
             return None
