@@ -23,7 +23,7 @@ import time
 from typing import Optional
 from uuid import UUID, uuid4
 
-from mockito import expect, mock, unstub, when
+from mockito import ANY, expect, mock, unstub, when
 import pytest
 import requests
 from requests import HTTPError, Response
@@ -95,9 +95,15 @@ def sample_circuit():
         'name': 'The circuit',
         'instructions': [
             {'name': 'cz', 'qubits': ['Qubit A', 'Qubit B'], 'args': {}},
-            {'name': 'phased_rx', 'qubits': ['Qubit A'], 'args': {'phase_t': 0.7, 'angle_t': 0.25}},
+            {
+                'name': 'phased_rx',
+                'implementation': 'drag_gaussian',
+                'qubits': ['Qubit A'],
+                'args': {'phase_t': 0.7, 'angle_t': 0.25},
+            },
             {'name': 'measurement', 'qubits': ['Qubit A'], 'args': {'output_label': 'A'}},
         ],
+        'metadata': {'experiment_type': 'test', 'qubits': (0, 1), 'values': [0.01686514, 0.05760602]},
     }
 
 
@@ -114,9 +120,10 @@ def sample_quantum_architecture():
 
 
 class MockJsonResponse:
-    def __init__(self, status_code: int, json_data: dict):
+    def __init__(self, status_code: int, json_data: dict, history: Optional[list[Response]] = None):
         self.status_code = status_code
         self.json_data = json_data
+        self.history = history
 
     @property
     def text(self):
@@ -137,14 +144,23 @@ def generate_server_stubs(base_url, sample_circuit):
     when(requests).post(f'{base_url}/jobs', ...).thenReturn(MockJsonResponse(201, {'id': str(existing_run)}))
 
     when(requests).get(f'{base_url}/jobs/{existing_run}', ...).thenReturn(
-        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': [sample_circuit]}})
+        MockJsonResponse(
+            200, {'status': 'pending', 'metadata': {'request': {'shots': 42, 'circuits': [sample_circuit]}}}
+        )
     ).thenReturn(
         MockJsonResponse(
             200,
             {
                 'status': 'ready',
                 'measurements': [{'result': [[1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 1, 1], [1, 0, 1, 1]]}],
-                'metadata': {'shots': 42, 'circuits': [sample_circuit], 'calibration_set_id': calibration_set_id_value},
+                'metadata': {
+                    'calibration_set_id': calibration_set_id_value,
+                    'request': {
+                        'shots': 42,
+                        'circuits': [sample_circuit],
+                        'calibration_set_id': calibration_set_id_value,
+                    },
+                },
             },
         )
     )
@@ -236,11 +252,31 @@ def expect_status_request(url: str, access_token: Optional[str], times: int = 1)
     job_id = uuid4()
     headers = None if access_token is None else {'Authorization': f'Bearer {access_token}'}
     expect(requests, times=times).get(f'{url}/jobs/{job_id}', headers=headers, timeout=REQUESTS_TIMEOUT).thenReturn(
-        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': []}})
+        MockJsonResponse(200, {'status': 'pending', 'metadata': {'request': {'shots': 42, 'circuits': []}}})
     )
     return job_id
 
 
+def expect_submit_circuits_request(
+    url: str, access_token: Optional[str], times: int = 1, response_status: int = 200
+) -> UUID:
+    """Prepare for submit_circuits request.
+
+    Args:
+        url: server URL for the status request
+        access_token: access token to use in Authorization header
+            If not set, expect request to have no Authorization header
+        times: number of times the status request is expected to be made
+        response_status: status code to return in the response
+    """
+    job_id = uuid4()
+    headers = None if access_token is None else {'Authorization': f'Bearer {access_token}', 'Expect': '100-Continue'}
+    expect(requests, times=times).post(
+        f'{url}/jobs', json=ANY(dict), headers=headers, timeout=REQUESTS_TIMEOUT
+    ).thenReturn(MockJsonResponse(response_status, {'id': str(job_id)}))
+    return job_id
+
+
 def expect_logout(auth_server_url: str, refresh_token: str):
     """Prepare for logout request.
 
